local band
band = require('bit').band
local floor
floor = math.floor
local CPU = require('CPU')
local Mem = require('Mem')
local CIA1 = require('CIA.CIA1')
local CIA2 = require('CIA.CIA2')
local VIC = require('VIC')
local SID = require('SID')
local portSwitch
portSwitch = function(self, val)
  local f = band(val, 0x07)
  local _exp_0 = f
  if 0 == _exp_0 or 4 == _exp_0 then
    self:switchRAM(0x0000, 0xffff)
    self:setIO(false)
  elseif 1 == _exp_0 then
    self:switchRAM(0x0000, 0xcfff)
    self:switchCHAR()
    self:switchRAM(0xe000, 0xffff)
    self:setIO(false)
  elseif 2 == _exp_0 then
    self:switchRAM(0x0000, 0xcfff)
    self:switchCHAR()
    self:switchKERNAL()
    self:setIO(false)
  elseif 3 == _exp_0 then
    self:switchRAM(0x0000, 0x9fff)
    self:switchBASIC()
    self:switchRAM(0xc000, 0xdfff)
    self:switchKERNAL()
    self:setIO(false)
  elseif 5 == _exp_0 then
    self:switchRAM(0x0000, 0xcfff)
    self:switchIO()
    self:switchRAM(0xe000, 0xffff)
    self:setIO(true)
  elseif 6 == _exp_0 then
    self:switchRAM(0x0000, 0xcfff)
    self:switchIO()
    self:switchKERNAL()
    self:setIO(true)
  elseif 7 == _exp_0 then
    self:switchRAM(0x0000, 0x9fff)
    self:switchBASIC()
    self:switchRAM(0xc000, 0xcfff)
    self:switchIO()
    self:switchKERNAL()
    self:setIO(true)
  end
  return val
end
local C64
do
  local _class_0
  local _base_0 = {
    NTSC = 1.023,
    PAL = .985,
    TEST = 200,
    getMem = function(self)
      return self._mem
    end,
    getCPU = function(self)
      return self._cpu
    end,
    getVIC = function(self)
      return self._vic
    end,
    getCIA1 = function(self)
      return self._cia1
    end,
    getFreq = function(self)
      return self._freq
    end,
    reset = function(self)
      return self._cpu:reset()
    end,
    addDevice = function(self, lo, hi, device)
      for i = lo, hi do
        self._devices[i % 0x1000] = device
      end
    end,
    tick = function(self, dt)
      local steps = floor(self._freq * dt + .5)
      if CPU_TEST then
        local cpu = self._cpu
        for i = 1, steps do
          cpu:step()
        end
      else
        local cpu, cia1, cia2, vic, sid = self._cpu, self._cia1, self._cia2, self._vic, self._sid
        for i = 1, steps do
          cia1:tick()
          cia2:tick()
          sid:tick()
          cpu._cycles = cpu._cycles + vic:tick()
          cpu:step()
        end
      end
      return steps
    end,
    _ioRead = function(self, addr)
      return self._devices[addr % 0x1000]:read(addr)
    end,
    _ioWrite = function(self, addr, value)
      return self._devices[addr % 0x1000]:write(addr, value)
    end
  }
  _base_0.__index = _base_0
  _class_0 = setmetatable({
    __init = function(self, _system)
      if _system == nil then
        _system = CPU_TEST and 'TEST' or 'PAL'
      end
      self._system = _system
      self._freq = floor(1e6 * C64[self._system] + .5)
      local portRead
      portRead = function(addr)
        return self._cpu:readPort(addr)
      end
      local portWrite
      portWrite = function(addr, val)
        val = self._cpu:writePort(addr, val)
        if addr == 1 then
          return portSwitch(self._mem, val)
        end
      end
      local ioRead
      ioRead = function(addr)
        return self:_ioRead(addr)
      end
      local ioWrite
      ioWrite = function(addr, value)
        return self:_ioWrite(addr, value)
      end
      local sramDevice = {
        read = function(_, addr)
          return self._mem.io[band(addr, 0xfff)]
        end,
        write = function(_, addr, val)
          self._mem.io[band(addr, 0xfff)] = val
          return val
        end
      }
      self._devices = { }
      for i = 0x800, 0xfff do
        self._devices[i] = sramDevice
      end
      self._mem = Mem(portRead, portWrite, ioRead, ioWrite)
      self._cpu = CPU(self._mem)
      self._cia1, self._cia2 = CIA1(self), CIA2(self)
      self._vic = VIC(self)
      self._sid = SID(self)
    end,
    __base = _base_0,
    __name = "C64"
  }, {
    __index = _base_0,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  C64 = _class_0
  return _class_0
end
