local ffi = require('ffi')
local bor, band, bnot, bxor, lshift, rshift
do
  local _obj_0 = require('bit')
  bor, band, bnot, bxor, lshift, rshift = _obj_0.bor, _obj_0.band, _obj_0.bnot, _obj_0.bxor, _obj_0.lshift, _obj_0.rshift
end
local min, max
do
  local _obj_0 = math
  min, max = _obj_0.min, _obj_0.max
end
local read8
read8 = function(self, addr)
  return self:read8(addr)
end
local writea
writea = function(self, addr, v)
  self.A = v
end
local write8
write8 = function(self, addr, v)
  return self:write8(addr, v)
end
local accum
accum = function(self)
  return self.A
end
local penalty
penalty = function(base, addr)
  return bxor(rshift(base, 8), rshift(addr, 8)) % 2
end
local tou8, tos8, tou16
do
  local tmp = ffi.new([[		struct {
			uint8_t  u8;
			int8_t   s8;
			uint16_t u16;
		}
	]])
  local u8
  u8 = function(self)
    tmp.u8 = self
    return tmp.u8
  end
  local s8
  s8 = function(self)
    tmp.s8 = self
    return tmp.s8
  end
  local u16
  u16 = function(self)
    tmp.u16 = self
    return tmp.u16
  end
  tou8, tos8, tou16 = u8, s8, u16
end
local imm
imm = function(self)
  return self.PC
end
local zp
zp = function(self)
  return self:read8(self.PC)
end
local zpx
zpx = function(self)
  return tou8(self:read8(self.PC) + self.X)
end
local zpy
zpy = function(self)
  return tou8(self:read8(self.PC) + self.Y)
end
local abs
abs = function(self)
  return self:read16(self.PC)
end
local absx
absx = function(self)
  return tou16(self:read16(self.PC) + self.X)
end
local absy
absy = function(self)
  return tou16(self:read16(self.PC) + self.Y)
end
local rel
rel = function(self)
  return tos8(self:read8(self.PC))
end
local absxp
absxp = function(self)
  local base = self:read16(self.PC)
  local addr = tou16(base + self.X)
  self._cycles = self._cycles + penalty(base, addr)
  return addr
end
local absyp
absyp = function(self)
  local base = self:read16(self.PC)
  local addr = tou16(base + self.Y)
  self._cycles = self._cycles + penalty(base, addr)
  return addr
end
local ind
ind = function(self)
  local b1 = self:read8(self.PC)
  local b2 = tou8(self.PC) == 0xff and self:read8(band(self.PC, 0xff00)) or self:read8(tou16(self.PC + 1))
  return self:read16(b1 + b2 * 0x100)
end
local indx
indx = function(self)
  local b1 = tou8(self:read8(self.PC) + self.X)
  return self:read8(b1) + self:read8(tou8(b1 + 1)) * 0x100
end
local indy
indy = function(self)
  local b1 = self:read8(self.PC)
  return tou16((self:read8(b1) + self:read8(tou8(b1 + 1)) * 0x100 + self.Y))
end
local indyp
indyp = function(self)
  local b1 = self:read8(self.PC)
  local base = self:read8(b1) + self:read8(tou8(b1 + 1)) * 0x100
  local addr = tou16((base + self.Y))
  self._cycles = self._cycles + penalty(base, addr)
  return addr
end
local NOP
NOP = function(self) end
local CLC
CLC = function(self)
  return self:setC(0x00)
end
local CLD
CLD = function(self)
  return self:setD(0x00)
end
local CLI
CLI = function(self)
  return self:setI(0x00)
end
local CLV
CLV = function(self)
  return self:setV(0x00)
end
local SEC
SEC = function(self)
  return self:setC(0x01)
end
local SED
SED = function(self)
  return self:setD(0x08)
end
local SEI
SEI = function(self)
  return self:setI(0x04)
end
local PHA
PHA = function(self)
  return self:push8(self.A)
end
local PHP
PHP = function(self)
  return self:push8(bor(self.P, 0x10))
end
local TAX
TAX = function(self)
  self.X = self:updateZN(self.A)
end
local TAY
TAY = function(self)
  self.Y = self:updateZN(self.A)
end
local TXA
TXA = function(self)
  self.A = self:updateZN(self.X)
end
local TYA
TYA = function(self)
  self.A = self:updateZN(self.Y)
end
local TSX
TSX = function(self)
  self.X = self:updateZN(self.SP)
end
local TXS
TXS = function(self)
  self.SP = self.X
end
local INX
INX = function(self)
  self.X = self:updateZN(tou8(self.X + 1))
end
local DEX
DEX = function(self)
  self.X = self:updateZN(tou8(self.X - 1))
end
local INY
INY = function(self)
  self.Y = self:updateZN(tou8(self.Y + 1))
end
local DEY
DEY = function(self)
  self.Y = self:updateZN(tou8(self.Y - 1))
end
local PLA
PLA = function(self)
  self.A = self:updateZN(self:pop8())
end
local PLP
PLP = function(self)
  self.P = bor(self:pop8(), 0x20)
end
local RTI
RTI = function(self)
  self.P, self.PC = self:pop8(), self:pop16()
end
local RTS
RTS = function(self)
  self.PC = self:pop16() + 1
end
local ORA
ORA = function(am)
  return function(self)
    self.A = self:updateZN(bor(self.A, self:read8(am(self))))
  end
end
local AND
AND = function(am)
  return function(self)
    self.A = self:updateZN(band(self.A, self:read8(am(self))))
  end
end
local EOR
EOR = function(am)
  return function(self)
    self.A = self:updateZN(bxor(self.A, self:read8(am(self))))
  end
end
local LDA
LDA = function(am)
  return function(self)
    self.A = self:updateZN(self:read8(am(self)))
  end
end
local LDX
LDX = function(am)
  return function(self)
    self.X = self:updateZN(self:read8(am(self)))
  end
end
local LDY
LDY = function(am)
  return function(self)
    self.Y = self:updateZN(self:read8(am(self)))
  end
end
local STA
STA = function(am)
  return function(self)
    return self:write8(am(self), self.A)
  end
end
local STX
STX = function(am)
  return function(self)
    return self:write8(am(self), self.X)
  end
end
local STY
STY = function(am)
  return function(self)
    return self:write8(am(self), self.Y)
  end
end
local BRK
BRK = function(self)
  self:push16(self.PC + 1)
  self:push8(bor(self.P, 0x10))
  self:setI(0x04)
  self.PC = self:read16(0xfffe)
end
local JSR
JSR = function(self)
  self:push16(tou16(self.PC + 1))
  self.PC = self:read16(self.PC) - 2
end
local ASL
ASL = function(am, r, w)
  return function(self)
    local addr = am(self)
    local v = lshift(r(self, addr), 1)
    self:setC(rshift(v, 8))
    local a = self:updateZN(tou8(v))
    return w(self, addr, a)
  end
end
local BIT
BIT = function(am)
  return function(self)
    local v = self:read8(am(self))
    self:updateZ(band(self.A, v))
    self.P = bor(self.P % 0x40, band(v, 0xc0))
  end
end
local ROL
ROL = function(am, r, w)
  return function(self)
    local addr = am(self)
    local a = r(self, addr)
    local v = bor(a * 2, self.P % 2)
    self:setC(rshift(v, 8))
    a = tou8(v)
    w(self, addr, a)
    self:updateZ(a)
    return self:updateN(v)
  end
end
local LSR
LSR = function(am, r, w)
  return function(self)
    local addr = am(self)
    local a = r(self, addr)
    local v = rshift(a, 1)
    self:setC(a % 2)
    w(self, addr, v)
    return self:updateZN(v)
  end
end
local ROR
ROR = function(am, r, w)
  return function(self)
    local addr = am(self)
    local a = r(self, addr)
    local v = a / 2 + self.P % 2 * 128
    self:setC(a % 2)
    a = tou8(v)
    w(self, addr, a)
    self:updateZ(a)
    return self:updateN(v)
  end
end
local ADC
ADC = function(am)
  return function(self)
    local m = self:read8(am(self))
    local res = self.A + m + self.P % 2
    self:updateV(res, self.A, m)
    if self:testStatus(0x08) then
      if self.A % 0x10 + m % 0x10 + self.P % 2 > 0x09 then
        res = res + 0x06
      end
      if res > 0x99 then
        res = res + 0x60
      end
      self:setC(max(0x00, min(res - 0x99, 0x01)))
    else
      self:setC(rshift(res, 8))
    end
    self.A = self:updateZN(tou8(res))
  end
end
local SBC
SBC = function(am)
  return function(self)
    local m = self:read8(am(self))
    local res = tou16(self.A - m - (1 - self.P % 2))
    self:updateV(self.A, res, m)
    if self:testStatus(0x08) then
      if self.A % 0x10 - (1 - self.P % 2) < m % 0x10 then
        res = res - 0x06
      end
      if res > 0x99 then
        res = res - 0x60
      end
    end
    self:setC(1 - band(rshift(res, 8), 1))
    self.A = self:updateZN(tou8(res))
  end
end
local CMP
CMP = function(am)
  return function(self)
    local val = self.A - self:read8(am(self))
    self:setC(max(0x00, min(val + 1, 0x01)))
    return self:updateZN(val)
  end
end
local CPX
CPX = function(am)
  return function(self)
    local val = self.X - self:read8(am(self))
    self:setC(max(0x00, min(val + 1, 0x01)))
    return self:updateZN(val)
  end
end
local CPY
CPY = function(am)
  return function(self)
    local val = self.Y - self:read8(am(self))
    self:setC(max(0x00, min(val + 1, 0x01)))
    return self:updateZN(val)
  end
end
local DEC
DEC = function(am)
  return function(self)
    local addr = am(self)
    local v = self:updateZN(tou8(self:read8(addr) - 1))
    return self:write8(addr, v)
  end
end
local INC
INC = function(am)
  return function(self)
    local addr = am(self)
    local v = self:updateZN(tou8(self:read8(addr) + 1))
    return self:write8(addr, v)
  end
end
local trap = CPU_TEST and function(self)
  return error(("trapped: %04x @ %d"):format(self.PC - 1, tonumber(self._instCounter)))
end
local BRANCH
BRANCH = function(flag, ifset)
  if CPU_TEST then
    return function(self)
      if not (self:testStatus(flag) == ifset) then
        return 
      end
      local op = rel(self)
      if op == -2 then
        trap(self)
      end
      local addr = tou16(self.PC + op)
      self._cycles = self._cycles + (1 + penalty(self.PC, addr))
      self.PC = addr
    end
  else
    return function(self)
      if not (self:testStatus(flag) == ifset) then
        return 
      end
      local addr = tou16(self.PC + rel(self))
      self._cycles = self._cycles + (1 + penalty(self.PC, addr))
      self.PC = addr
    end
  end
end
local JMP
JMP = function(am)
  if CPU_TEST then
    return function(self)
      local addr = am(self)
      if addr == self.PC - 1 then
        if addr == 0x3469 then
          CPU_TEST()
        else
          trap(self)
        end
      end
      self.PC = addr - 2
    end
  else
    return function(self)
      self.PC = am(self) - 2
    end
  end
end
local opcodes = {
  [0x00] = {
    1,
    7,
    BRK
  },
  [0x01] = {
    2,
    6,
    ORA(indx)
  },
  [0x05] = {
    2,
    3,
    ORA(zp)
  },
  [0x06] = {
    2,
    5,
    ASL(zp, read8, write8)
  },
  [0x08] = {
    1,
    3,
    PHP
  },
  [0x09] = {
    2,
    2,
    ORA(imm)
  },
  [0x0a] = {
    1,
    2,
    ASL(accum, accum, writea)
  },
  [0x0d] = {
    3,
    4,
    ORA(abs)
  },
  [0x0e] = {
    3,
    6,
    ASL(abs, read8, write8)
  },
  [0x10] = {
    2,
    2,
    BRANCH(0x80, false)
  },
  [0x11] = {
    2,
    5,
    ORA(indyp)
  },
  [0x15] = {
    2,
    4,
    ORA(zpx)
  },
  [0x16] = {
    2,
    6,
    ASL(zpx, read8, write8)
  },
  [0x18] = {
    1,
    2,
    CLC
  },
  [0x19] = {
    3,
    4,
    ORA(absyp)
  },
  [0x1d] = {
    3,
    4,
    ORA(absxp)
  },
  [0x1e] = {
    3,
    7,
    ASL(absx, read8, write8)
  },
  [0x20] = {
    3,
    6,
    JSR
  },
  [0x21] = {
    2,
    6,
    AND(indx)
  },
  [0x24] = {
    2,
    3,
    BIT(zp)
  },
  [0x25] = {
    2,
    3,
    AND(zp)
  },
  [0x26] = {
    2,
    5,
    ROL(zp, read8, write8)
  },
  [0x28] = {
    1,
    4,
    PLP
  },
  [0x29] = {
    2,
    2,
    AND(imm)
  },
  [0x2a] = {
    1,
    2,
    ROL(accum, accum, writea)
  },
  [0x2c] = {
    3,
    4,
    BIT(abs)
  },
  [0x2d] = {
    3,
    4,
    AND(abs)
  },
  [0x2e] = {
    3,
    6,
    ROL(abs, read8, write8)
  },
  [0x30] = {
    2,
    2,
    BRANCH(0x80, true)
  },
  [0x31] = {
    2,
    5,
    AND(indyp)
  },
  [0x35] = {
    2,
    4,
    AND(zpx)
  },
  [0x36] = {
    2,
    6,
    ROL(zpx, read8, write8)
  },
  [0x38] = {
    1,
    2,
    SEC
  },
  [0x39] = {
    3,
    4,
    AND(absyp)
  },
  [0x3d] = {
    3,
    4,
    AND(absxp)
  },
  [0x3e] = {
    3,
    7,
    ROL(absx, read8, write8)
  },
  [0x40] = {
    1,
    6,
    RTI
  },
  [0x41] = {
    2,
    6,
    EOR(indx)
  },
  [0x45] = {
    2,
    3,
    EOR(zp)
  },
  [0x46] = {
    2,
    5,
    LSR(zp, read8, write8)
  },
  [0x48] = {
    1,
    3,
    PHA
  },
  [0x49] = {
    2,
    2,
    EOR(imm)
  },
  [0x4a] = {
    1,
    2,
    LSR(accum, accum, writea)
  },
  [0x4c] = {
    3,
    3,
    JMP(abs)
  },
  [0x4d] = {
    3,
    4,
    EOR(abs)
  },
  [0x4e] = {
    3,
    6,
    LSR(abs, read8, write8)
  },
  [0x50] = {
    2,
    2,
    BRANCH(0x40, false)
  },
  [0x51] = {
    2,
    5,
    EOR(indyp)
  },
  [0x55] = {
    2,
    4,
    EOR(zpx)
  },
  [0x56] = {
    2,
    6,
    LSR(zpx, read8, write8)
  },
  [0x58] = {
    1,
    2,
    CLI
  },
  [0x59] = {
    3,
    4,
    EOR(absyp)
  },
  [0x5d] = {
    3,
    4,
    EOR(absxp)
  },
  [0x5e] = {
    3,
    7,
    LSR(absx, read8, write8)
  },
  [0x60] = {
    1,
    6,
    RTS
  },
  [0x61] = {
    2,
    6,
    ADC(indx)
  },
  [0x65] = {
    2,
    3,
    ADC(zp)
  },
  [0x66] = {
    2,
    5,
    ROR(zp, read8, write8)
  },
  [0x68] = {
    1,
    4,
    PLA
  },
  [0x69] = {
    2,
    2,
    ADC(imm)
  },
  [0x6a] = {
    1,
    2,
    ROR(accum, accum, writea)
  },
  [0x6c] = {
    3,
    5,
    JMP(ind)
  },
  [0x6d] = {
    3,
    4,
    ADC(abs)
  },
  [0x6e] = {
    3,
    6,
    ROR(abs, read8, write8)
  },
  [0x70] = {
    2,
    2,
    BRANCH(0x40, true)
  },
  [0x71] = {
    2,
    5,
    ADC(indyp)
  },
  [0x75] = {
    2,
    4,
    ADC(zpx)
  },
  [0x76] = {
    2,
    6,
    ROR(zpx, read8, write8)
  },
  [0x78] = {
    1,
    2,
    SEI
  },
  [0x79] = {
    3,
    4,
    ADC(absyp)
  },
  [0x7d] = {
    3,
    4,
    ADC(absxp)
  },
  [0x7e] = {
    3,
    7,
    ROR(absx, read8, write8)
  },
  [0x81] = {
    2,
    6,
    STA(indx)
  },
  [0x84] = {
    2,
    3,
    STY(zp)
  },
  [0x85] = {
    2,
    3,
    STA(zp)
  },
  [0x86] = {
    2,
    3,
    STX(zp)
  },
  [0x88] = {
    1,
    2,
    DEY
  },
  [0x8a] = {
    1,
    2,
    TXA
  },
  [0x8c] = {
    3,
    4,
    STY(abs)
  },
  [0x8d] = {
    3,
    4,
    STA(abs)
  },
  [0x8e] = {
    3,
    4,
    STX(abs)
  },
  [0x90] = {
    2,
    2,
    BRANCH(0x01, false)
  },
  [0x91] = {
    2,
    6,
    STA(indy)
  },
  [0x94] = {
    2,
    4,
    STY(zpx)
  },
  [0x95] = {
    2,
    4,
    STA(zpx)
  },
  [0x96] = {
    2,
    4,
    STX(zpy)
  },
  [0x98] = {
    1,
    2,
    TYA
  },
  [0x99] = {
    3,
    5,
    STA(absy)
  },
  [0x9a] = {
    1,
    2,
    TXS
  },
  [0x9d] = {
    3,
    5,
    STA(absx)
  },
  [0xa0] = {
    2,
    2,
    LDY(imm)
  },
  [0xa1] = {
    2,
    6,
    LDA(indx)
  },
  [0xa2] = {
    2,
    2,
    LDX(imm)
  },
  [0xa4] = {
    2,
    3,
    LDY(zp)
  },
  [0xa5] = {
    2,
    3,
    LDA(zp)
  },
  [0xa6] = {
    2,
    3,
    LDX(zp)
  },
  [0xa8] = {
    1,
    2,
    TAY
  },
  [0xa9] = {
    2,
    2,
    LDA(imm)
  },
  [0xaa] = {
    1,
    2,
    TAX
  },
  [0xac] = {
    3,
    4,
    LDY(abs)
  },
  [0xad] = {
    3,
    4,
    LDA(abs)
  },
  [0xae] = {
    3,
    4,
    LDX(abs)
  },
  [0xb0] = {
    2,
    2,
    BRANCH(0x01, true)
  },
  [0xb1] = {
    2,
    5,
    LDA(indyp)
  },
  [0xb4] = {
    2,
    4,
    LDY(zpx)
  },
  [0xb5] = {
    2,
    4,
    LDA(zpx)
  },
  [0xb6] = {
    2,
    4,
    LDX(zpy)
  },
  [0xb8] = {
    1,
    2,
    CLV
  },
  [0xb9] = {
    3,
    4,
    LDA(absyp)
  },
  [0xba] = {
    1,
    2,
    TSX
  },
  [0xbc] = {
    3,
    4,
    LDY(absxp)
  },
  [0xbd] = {
    3,
    4,
    LDA(absxp)
  },
  [0xbe] = {
    3,
    4,
    LDX(absyp)
  },
  [0xc0] = {
    2,
    2,
    CPY(imm)
  },
  [0xc1] = {
    2,
    6,
    CMP(indx)
  },
  [0xc4] = {
    2,
    3,
    CPY(zp)
  },
  [0xc5] = {
    2,
    3,
    CMP(zp)
  },
  [0xc6] = {
    2,
    5,
    DEC(zp)
  },
  [0xc8] = {
    1,
    2,
    INY
  },
  [0xc9] = {
    2,
    2,
    CMP(imm)
  },
  [0xca] = {
    1,
    2,
    DEX
  },
  [0xcc] = {
    3,
    4,
    CPY(abs)
  },
  [0xcd] = {
    3,
    4,
    CMP(abs)
  },
  [0xce] = {
    3,
    6,
    DEC(abs)
  },
  [0xd0] = {
    2,
    2,
    BRANCH(0x02, false)
  },
  [0xd1] = {
    2,
    5,
    CMP(indyp)
  },
  [0xd5] = {
    2,
    4,
    CMP(zpx)
  },
  [0xd6] = {
    2,
    6,
    DEC(zpx)
  },
  [0xd8] = {
    1,
    2,
    CLD
  },
  [0xd9] = {
    3,
    4,
    CMP(absyp)
  },
  [0xdd] = {
    3,
    4,
    CMP(absxp)
  },
  [0xde] = {
    3,
    7,
    DEC(absx)
  },
  [0xe0] = {
    2,
    2,
    CPX(imm)
  },
  [0xe1] = {
    2,
    6,
    SBC(indx)
  },
  [0xe4] = {
    2,
    3,
    CPX(zp)
  },
  [0xe5] = {
    2,
    3,
    SBC(zp)
  },
  [0xe6] = {
    2,
    5,
    INC(zp)
  },
  [0xe8] = {
    1,
    2,
    INX
  },
  [0xe9] = {
    2,
    2,
    SBC(imm)
  },
  [0xea] = {
    1,
    2,
    NOP
  },
  [0xec] = {
    3,
    4,
    CPX(abs)
  },
  [0xed] = {
    3,
    4,
    SBC(abs)
  },
  [0xee] = {
    3,
    6,
    INC(abs)
  },
  [0xf0] = {
    2,
    2,
    BRANCH(0x02, true)
  },
  [0xf1] = {
    2,
    5,
    SBC(indyp)
  },
  [0xf5] = {
    2,
    4,
    SBC(zpx)
  },
  [0xf6] = {
    2,
    6,
    INC(zpx)
  },
  [0xf8] = {
    1,
    2,
    SED
  },
  [0xf9] = {
    3,
    4,
    SBC(absyp)
  },
  [0xfd] = {
    3,
    4,
    SBC(absxp)
  },
  [0xfe] = {
    3,
    7,
    INC(absx)
  }
}
for i = 0, 0xff do
  opcodes[i] = opcodes[i] or opcodes[0xea]
end
return opcodes
