local ffi = require('ffi')
local band, bor, bxor, lshift, rshift
do
  local _obj_0 = require('bit')
  band, bor, bxor, lshift, rshift = _obj_0.band, _obj_0.bor, _obj_0.bxor, _obj_0.lshift, _obj_0.rshift
end
local min, max
do
  local _obj_0 = math
  min, max = _obj_0.min, _obj_0.max
end
local M0X, M0Y = 0x00, 0x01
local M1X, M1Y = 0x02, 0x03
local M2X, M2Y = 0x04, 0x05
local M3X, M3Y = 0x06, 0x07
local M4X, M4Y = 0x08, 0x09
local M5X, M5Y = 0x0a, 0x0b
local M6X, M6Y = 0x0c, 0x0d
local M7X, M7Y = 0x0e, 0x0f
local MXMSB = 0x10
local CR1 = 0x11
local RST = 0x12
local LPX = 0x13
local LPY = 0x14
local MEN = 0x15
local CR2 = 0x16
local MYE = 0x17
local MPTR = 0x18
local IR = 0x19
local IE = 0x1a
local MDP = 0x1b
local MMC = 0x1c
local MXE = 0x1d
local MM = 0x1e
local MD = 0x1f
local EC = 0x20
local B0C = 0x21
local B1C = 0x22
local B2C = 0x23
local B3C = 0x24
local MM0 = 0x25
local MM1 = 0x26
local M0C = 0x27
local M1C = 0x28
local M2C = 0x29
local M3C = 0x2a
local M4C = 0x2b
local M5C = 0x2c
local M6C = 0x2d
local M7C = 0x2e
local VIC_cdef = [[	struct __attribute__((__packed__)) {
		uint16_t m[8];      // sprite pos
		uint8_t mmsb;       // sprite x msb
		uint8_t cr1;        // control reg 1
		uint8_t rst;        // raster
		uint8_t lpx, lpy;   // light pen
		uint8_t me;         // sprite enable
		uint8_t cr2;        // control reg 2
		uint8_t mye;        // sprite y expansion
		uint8_t mptr;       // memory pointers
		uint8_t ir;         // interrupt register
		uint8_t ie;         // interrupt enable
		uint8_t mdp;        // sprite data priority
		uint8_t mmc;        // sprite multicolor
		uint8_t mxe;        // sprite x expansion
		uint8_t mmc, mdc;   // sprite collision
		uint8_t ec;         // border
		uint8_t bc[4];      // background
		uint8_t mm[2];      // sprite multicolor
		uint8_t mc[8];      // sprite color
	}
]]
local RASTERLEN, LINES = 504, 312
local STARTX = math.floor((RASTERLEN - SCREEN_WIDTH) * .5)
STARTX = STARTX - (STARTX % 8)
local STARTY = math.floor((LINES - SCREEN_HEIGHT) * .5)
local ENDX, ENDY = STARTX + SCREEN_WIDTH, STARTY + SCREEN_HEIGHT
local BSTARTX = math.floor((SCREEN_WIDTH - 320) * .5)
BSTARTX = BSTARTX - (BSTARTX % 8)
local BSTARTY = math.floor((SCREEN_HEIGHT - 200) * .5)
local BENDX, BENDY = BSTARTX + 320, BSTARTY + 200
local steps = 0
local VIC
VIC = function(system)
  local mem = system:getMem()
  local cpu = system:getCPU()
  local framebuffer = ffi.new('uint8_t[?]', SCREEN_WIDTH * SCREEN_HEIGHT)
  local crom = mem.crom
  local cram = ffi.cast('uint8_t*', mem.io + 0x800)
  local sram = ffi.cast('uint8_t*', mem.ram + 0x400)
  local regs, mode = nil, nil
  local cycles, line, rasterPos = 0, 0, 0
  local hblank, vblank = true, true
  return ffi.metatype(ffi.typeof(VIC_cdef), {
    __new = function(self)
      local vic = ffi.new(self)
      regs = ffi.cast('char*', vic)
      system:addDevice(0xd000, 0xd3ff, vic)
      return vic
    end,
    __index = {
      render = function(self, target)
        local t = ffi.cast('char*', target:getFFIPointer())
        ffi.copy(t, framebuffer, ffi.sizeof(framebuffer))
        return target
      end,
      read = function(self, addr)
        local reg = band(addr, 0xf03f) - 0xd000
        return regs[reg]
      end,
      write = function(self, addr, val)
        local reg = band(addr, 0xf03f) - 0xd000
        regs[reg] = val
        return val
      end,
      tick = function(self)
        cycles = 0
        if not (hblank or vblank) then
          mode = rshift(band(self.cr1, 0x20), 5) + rshift(band(self.cr1, 0x40), 5) + rshift(band(self.cr2, 0x10), 3)
          self._modes[mode + 1](self)
        end
        self:_updateRST()
        return cycles
      end,
      _updateRST = function(self)
        rasterPos = rasterPos + 8
        line = self.rst + lshift(band(self.cr1, 0x80), 1)
        if rasterPos >= RASTERLEN then
          rasterPos = 0
          line = (line + 1) % LINES
        end
        self.rst = band(line, 0xff)
        self.cr1 = bor(band(self.cr1, 0x7f), rshift(band(line, 0x100), 1))
        hblank = rasterPos < STARTX or rasterPos >= ENDX
        vblank = line < STARTY or line >= ENDY
      end,
      _modes = {
        function(self)
          local x, y = rasterPos - STARTX, line - STARTY
          local isBorder = x < BSTARTX or x >= BENDX or y < BSTARTY or y >= BENDY
          local ec, bc = band(self.ec, 0x0f), band(self.bc[0], 0x0f)
          if isBorder then
            local n = -max(-8, x - SCREEN_WIDTH)
            return ffi.fill(framebuffer + y * SCREEN_WIDTH + x, n, ec)
          else
            local sx, sy = x - BSTARTX, y - BSTARTY
            local cx, cy = rshift(sx, 3), band(sy, 0x07)
            local charoffs = rshift(sy, 3) * 0x28 + cx
            local char, fg = sram[charoffs], cram[charoffs]
            local tile = crom[char * 8 + cy]
            for p = 0, 7 do
              local b = rshift(band(tile, lshift(1, 7 - p)), 7 - p)
              framebuffer[y * SCREEN_WIDTH + x + p] = b * fg + (1 - b) * bc
            end
          end
        end,
        function(self) end,
        function(self) end,
        function(self) end,
        function(self) end,
        function(self) end,
        function(self) end,
        function(self) end
      },
      _rasterLine = function(self) end
    }
  })()
end
return function(system)
  return VIC(system)
end
