local ffi = require('ffi')
local graphics
graphics = love.graphics
local C64 = require('C64')
local system = C64()
local palette = love.graphics.newImage('assets/palette.png')
palette:setFilter('nearest')
if CPU_TEST then
  local tests = love.filesystem.read('6502_functional_test.bin')
  do
    system:getMem():load(tests)
    system:getCPU():reset(0x0400)
  end
else
  system:tick(2.2)
end
local screen = love.image.newImageData(SCREEN_WIDTH, SCREEN_HEIGHT, 'r8')
local screenImg = graphics.newImage(screen)
screenImg:setFilter('nearest')
local screenCanvas = graphics.newCanvas(SCREEN_WIDTH, SCREEN_HEIGHT)
screenCanvas:setFilter('nearest')
local scanlines, scanQuad
do
  local intensity = .25
  local sl = love.image.newImageData(1, 2)
  sl:setPixel(0, 0, 1, 1, 1, 1)
  sl:setPixel(0, 1, 1 - intensity, 1 - intensity, 1 - intensity, 1 - intensity)
  local img = graphics.newImage(sl)
  img:setWrap('repeat')
  local sw, sh = graphics.getDimensions()
  local quad = graphics.newQuad(0, 0, sw, sh, sl:getDimensions())
  scanlines, scanQuad = img, quad
end
local palShader = graphics.newShader([[	uniform Image pal;

	vec4 effect(vec4 col, Image tex, vec2 uv, vec2 fc) {
		float t = Texel(tex, uv).r * 16.0;
		vec4 c = Texel(pal, vec2(t + 1.0 / 32.0, .5));
		return col * c;
	}
]])
palShader:send('pal', palette)
local outputShader = graphics.newShader([[	vec4 effect(vec4 col, Image tex, vec2 uv, vec2 fc) {
		float r = Texel(tex, uv + .0015).r;
		float g = Texel(tex, uv).g;
		float b = Texel(tex, uv - .001).b;

		return vec4(r, g, b, 1.0);
	}
]])
local memid = love.image.newImageData(256, 256, 'r8')
local memimg = love.graphics.newImage(memid)
memimg:setFilter('nearest')
local time = 0
local steps = 0
love.update = function(dt)
  dt = math.min(1 / 20, dt)
  time = time + dt
  steps = system:tick(dt)
end
love.draw = function()
  graphics.line(0, 0, 0, 0)
  screenImg:replacePixels(system:getVIC():render(screen))
  do
    graphics.setCanvas(screenCanvas)
    graphics.setShader(palShader)
    graphics.draw(screenImg)
    graphics.setShader()
    graphics.setCanvas()
  end
  do
    local sx, sy = graphics.getDimensions()
    local scaleX, scaleY = sx / screen:getWidth(), sy / screen:getHeight()
    graphics.setShader(outputShader)
    graphics.draw(screenCanvas, 0, 0, 0, scaleX, scaleY)
    graphics.setShader()
    graphics.setBlendMode('multiply', 'premultiplied')
    graphics.draw(scanlines, scanQuad)
    graphics.setBlendMode('alpha', 'alphamultiply')
  end
  return love.window.setTitle(("LC64 | %.3f MHz | %d fps"):format(steps / (love.timer.getDelta() * 1e6), love.timer.getFPS()))
end
love.keypressed = function(key, scan)
  if key == 'f11' then
    return system:reset()
  end
  if key == 'f12' then
    return love.event.quit()
  end
  return system:getCIA1():keypressed(key, scan)
end
love.keyreleased = function(key, scan)
  return system:getCIA1():keyreleased(key, scan)
end
